"""Kernels that extends sklearn.gaussian_processes.kernels by allowing for custom array modules to be used (e.g.
Numpy, Cupy, Torch, etc) and supporting for batched operations """

import numpy as np
from sklearn.gaussian_process import kernels
from sklearn.base import clone
import math


class BatchedKernel(kernels.Kernel):
    """ Base class for Batched kernels """
    def __add__(self, b):
        if not isinstance(b, BatchedKernel):
            return Sum(self, ConstantKernel(b))
        return Sum(self, b)

    def __radd__(self, b):
        if not isinstance(b, BatchedKernel):
            return Sum(ConstantKernel(b), self)
        return Sum(b, self)

    def __mul__(self, b):
        if not isinstance(b, BatchedKernel):
            return Product(self, ConstantKernel(b))
        return Product(self, b)

    def __rmul__(self, b):
        if not isinstance(b, BatchedKernel):
            return Product(ConstantKernel(b), self)
        return Product(b, self)
    
    def __pow__(self, b):
        return Exponentiation(self, b)
    
    def bind_to_array_module(self, xp):
        res = BatchedKernelWrapper(self)
        res.xp = xp
        return res


class BatchedKernelWrapper(kernels.Kernel):
    """ Wrapper around a BatchedKernel that binds the kernel to the provided array module and allows to use the kernel 
    with a scikit-learn like interface 

    Parameters
    ----------

    base_kernel : BatchedKernel
        Kernel to wrap
    """

    def __init__(self, base_kernel):
        self.base_kernel = base_kernel
        self.xp = None

    def __call__(self, *args, **kwargs):
        return self.base_kernel(self.xp, *args, **kwargs)

    def diag(self, *args, **kwargs):
        return self.base_kernel.diag(self.xp, *args, **kwargs)
    
    def is_stationary(self):
        return self.base_kernel.is_stationary()
    
    @property
    def theta(self):
        return self.base_kernel.theta
    
    @theta.setter
    def theta(self, theta):
        self.base_kernel.theta = theta

    @property
    def bounds(self):
        return self.base_kernel.bounds

    def clone_with_theta(self, theta):
        res = BatchedKernelWrapper(self.base_kernel.clone_with_theta(theta))
        res.xp = self.xp
        return res

    def __getattr__(self, name):
        return getattr(self.base_kernel, name)
    
    def __repr__(self):
        return self.base_kernel.__repr__()
    
    def __str__(self):
        return self.base_kernel.__str__()
    
    def get_params(self, deep=True):
        return self.base_kernel.get_params(deep=deep)

    def set_params(self, **params):
        self.base_kernel.set_params(**params)
        return self

    @property
    def hyperparameters(self):
        """Returns a list of all hyperparameter specifications."""
        return self.base_kernel.hyperparameters
    
    def __sklearn_clone__(self):
        base_clone = clone(self.base_kernel)
        res = BatchedKernelWrapper(base_clone)
        res.xp = self.xp
        return res


class Sum(BatchedKernel, kernels.Sum):
    """The `Sum` kernel takes two kernels :math:`k_1` and :math:`k_2`
    and combines them via

    .. math::
        k_{sum}(X, Y) = k_1(X, Y) + k_2(X, Y)

    Note that the `__add__` magic method is overridden, so
    `Sum(RBF(), RBF())` is equivalent to using the + operator
    with `RBF() + RBF()`.


    Read more in the :ref:`User Guide <gp_kernels>`.

    .. versionadded:: 0.18

    Parameters
    ----------
    k1 : Kernel
        The first base-kernel of the sum-kernel

    k2 : Kernel
        The second base-kernel of the sum-kernel

    Examples
    --------
    >>> from sklearn.datasets import make_friedman2
    >>> from sklearn.gaussian_process import GaussianProcessRegressor
    >>> from sklearn.gaussian_process.kernels import RBF, Sum, ConstantKernel
    >>> X, y = make_friedman2(n_samples=500, noise=0, random_state=0)
    >>> kernel = Sum(ConstantKernel(2), RBF())
    >>> gpr = GaussianProcessRegressor(kernel=kernel,
    ...         random_state=0).fit(X, y)
    >>> gpr.score(X, y)
    1.0
    >>> kernel
    1.41**2 + RBF(length_scale=1)
    """

    def __call__(self, xp, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        X : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object
            Left argument of the returned kernel k(X, Y)

        Y : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object,\
                default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            is evaluated instead.

        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the log of
            the kernel hyperparameter is computed.

        Returns
        -------
        K : ndarray of shape (*batch_shape, n_samples_X, n_samples_Y)
            Kernel k(X, Y)

        K_gradient : ndarray of shape (*batch_shape, n_samples_X, n_samples_X, n_dims),\
                optional
            The gradient of the kernel k(X, X) with respect to the log of the
            hyperparameter of the kernel. Only returned when `eval_gradient`
            is True.
        """
        if eval_gradient:
            K1, K1_gradient = self.k1(xp, X, Y, eval_gradient=True)
            K2, K2_gradient = self.k2(xp, X, Y, eval_gradient=True)
            return K1 + K2, xp.concatenate((K1_gradient, K2_gradient), axis=-1)
        else:
            return self.k1(xp, X, Y) + self.k2(xp, X, Y)
        
    def diag(self, xp, X):
        """Returns the diagonal of the kernel k(X, X).

        The result of this method is identical to `np.diag(self(X))`; however,
        it can be evaluated more efficiently since only the diagonal is
        evaluated.

        Parameters
        ----------
        X : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object
            Argument to the kernel.

        Returns
        -------
        K_diag : ndarray of shape (*batch_shape, n_samples_X,)
            Diagonal of kernel k(X, X)
        """
        return self.k1.diag(xp, X) + self.k2.diag(xp, X)


class Product(BatchedKernel, kernels.Product):
    """The `Product` kernel takes two kernels :math:`k_1` and :math:`k_2`
    and combines them via

    .. math::
        k_{prod}(X, Y) = k_1(X, Y) * k_2(X, Y)

    Note that the `__mul__` magic method is overridden, so
    `Product(RBF(), RBF())` is equivalent to using the * operator
    with `RBF() * RBF()`.

    Read more in the :ref:`User Guide <gp_kernels>`.

    .. versionadded:: 0.18

    ParametersX
    ----------
    k1 : Kernel
        The first base-kernel of the product-kernel

    k2 : Kernel
        The second base-kernel of the product-kernel


    Examples
    --------
    >>> from sklearn.datasets import make_friedman2
    >>> from sklearn.gaussian_process import GaussianProcessRegressor
    >>> from sklearn.gaussian_process.kernels import (RBF, Product,
    ...            ConstantKernel)
    >>> X, y = make_friedman2(n_samples=500, noise=0, random_state=0)
    >>> kernel = Product(ConstantKernel(2), RBF())
    >>> gpr = GaussianProcessRegressor(kernel=kernel,
    ...         random_state=0).fit(X, y)
    >>> gpr.score(X, y)
    1.0
    >>> kernel
    1.41**2 * RBF(length_scale=1)
    """

    def __call__(self, xp, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        X : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object
            Left argument of the returned kernel k(X, Y)

        Y : array-like of shape (*batch_shape, n_samples_Y, n_features) or list of object,\
            default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            is evaluated instead.

        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the log of
            the kernel hyperparameter is computed.

        Returns
        -------
        K : ndarray of shape (*batch_shape, n_samples_X, n_samples_Y)
            Kernel k(X, Y)

        K_gradient : ndarray of shape (*batch_shape, n_samples_X, n_samples_X, n_dims), \
                optional
            The gradient of the kernel k(X, X) with respect to the log of the
            hyperparameter of the kernel. Only returned when `eval_gradient`
            is True.
        """
        if eval_gradient:
            K1, K1_gradient = self.k1(xp, X, Y, eval_gradient=True)
            K2, K2_gradient = self.k2(xp, X, Y, eval_gradient=True)
            return K1 * K2, xp.concatenate(
                (K1_gradient * K2[..., None], K2_gradient * K1[..., None]), axis=-1)
        else:
            return self.k1(xp, X, Y) * self.k2(xp, X, Y)
        
    def diag(self, xp, X):
        """Returns the diagonal of the kernel k(X, X).

        The result of this method is identical to np.diag(self(X)); however,
        it can be evaluated more efficiently since only the diagonal is
        evaluated.

        Parameters
        ----------
        X : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object
            Argument to the kernel.

        Returns
        -------
        K_diag : ndarray of shape (*batch_shape, n_samples_X,)
            Diagonal of kernel k(X, X)
        """
        return self.k1.diag(xp, X) * self.k2.diag(xp, X)


class CompoundKernel(kernels.Kernel):
    """Kernel which is composed of a set of other kernels.

    .. versionadded:: 0.18

    Parameters
    ----------
    kernels : list of Kernels
        The other kernels

    Examples
    --------
    >>> from sklearn.gaussian_process.kernels import WhiteKernel
    >>> from sklearn.gaussian_process.kernels import RBF
    >>> from sklearn.gaussian_process.kernels import CompoundKernel
    >>> kernel = CompoundKernel(
    ...     [WhiteKernel(noise_level=3.0), RBF(length_scale=2.0)])
    >>> print(kernel.bounds)
    [[-11.51292546  11.51292546]
     [-11.51292546  11.51292546]]
    >>> print(kernel.n_dims)
    2
    >>> print(kernel.theta)
    [1.09861229 0.69314718]
    """

    def __init__(self, kernels):
        self.kernels = kernels

    def get_params(self, deep=True):
        """Get parameters of this kernel.

        Parameters
        ----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : dict
            Parameter names mapped to their values.
        """
        return dict(kernels=self.kernels)

    @property
    def theta(self):
        """Returns the (flattened, log-transformed) non-fixed hyperparameters.

        Note that theta are typically the log-transformed values of the
        kernel's hyperparameters as this representation of the search space
        is more amenable for hyperparameter search, as hyperparameters like
        length-scales naturally live on a log-scale.

        Returns
        -------
        theta : ndarray of shape (n_dims,)
            The non-fixed, log-transformed hyperparameters of the kernel
        """
        return np.hstack([kernel.theta for kernel in self.kernels])

    @theta.setter
    def theta(self, theta):
        """Sets the (flattened, log-transformed) non-fixed hyperparameters.

        Parameters
        ----------
        theta : array of shape (n_dims,)
            The non-fixed, log-transformed hyperparameters of the kernel
        """
        k_dims = self.k1.n_dims
        for i, kernel in enumerate(self.kernels):
            kernel.theta = theta[i * k_dims : (i + 1) * k_dims]

    @property
    def bounds(self):
        """Returns the log-transformed bounds on the theta.

        Returns
        -------
        bounds : array of shape (n_dims, 2)
            The log-transformed bounds on the kernel's hyperparameters theta
        """
        return np.vstack([kernel.bounds for kernel in self.kernels])

    def __call__(self, xp, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.

        Note that this compound kernel returns the results of all simple kernel
        stacked along an additional axis.

        Parameters
        ----------
        X : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object, \
            default=None
            Left argument of the returned kernel k(X, Y)

        Y : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object, \
            default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            is evaluated instead.

        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the log of the
            kernel hyperparameter is computed.

        Returns
        -------
        K : ndarray of shape (*batch_shape, n_samples_X, n_samples_Y, n_kernels)
            Kernel k(X, Y)

        K_gradient : ndarray of shape \
                (n_samples_X, n_samples_X, n_dims, n_kernels), optional
            The gradient of the kernel k(X, X) with respect to the log of the
            hyperparameter of the kernel. Only returned when `eval_gradient`
            is True.
        """
        if eval_gradient:
            K = []
            K_grad = []
            for kernel in self.kernels:
                K_single, K_grad_single = kernel(X, Y, eval_gradient)
                K.append(K_single)
                K_grad.append(K_grad_single[..., np.newaxis])
            return xp.dstack(K), xp.concatenate(K_grad, 3)
        else:
            return xp.dstack([kernel(xp, X, Y, eval_gradient) for kernel in self.kernels])

    def diag(self, xp, X):
        """Returns the diagonal of the kernel k(X, X).

        The result of this method is identical to `np.diag(self(X))`; however,
        it can be evaluated more efficiently since only the diagonal is
        evaluated.

        Parameters
        ----------
        X : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object
            Argument to the kernel.

        Returns
        -------
        K_diag : ndarray of shape (*batch_shape, n_samples_X, n_kernels)
            Diagonal of kernel k(X, X)
        """
        return xp.vstack([kernel.diag(xp, X) for kernel in self.kernels]).T


class Exponentiation(kernels.Exponentiation):
    """The Exponentiation kernel takes one base kernel and a scalar parameter
    :math:`p` and combines them via

    .. math::
        k_{exp}(X, Y) = k(X, Y) ^p

    Note that the `__pow__` magic method is overridden, so
    `Exponentiation(RBF(), 2)` is equivalent to using the ** operator
    with `RBF() ** 2`.


    Read more in the :ref:`User Guide <gp_kernels>`.

    .. versionadded:: 0.18

    Parameters
    ----------
    kernel : Kernel
        The base kernel

    exponent : float
        The exponent for the base kernel


    Examples
    --------
    >>> from sklearn.datasets import make_friedman2
    >>> from sklearn.gaussian_process import GaussianProcessRegressor
    >>> from sklearn.gaussian_process.kernels import (RationalQuadratic,
    ...            Exponentiation)
    >>> X, y = make_friedman2(n_samples=500, noise=0, random_state=0)
    >>> kernel = Exponentiation(RationalQuadratic(), exponent=2)
    >>> gpr = GaussianProcessRegressor(kernel=kernel, alpha=5,
    ...         random_state=0).fit(X, y)
    >>> gpr.score(X, y)
    0.419...
    >>> gpr.predict(X[:1,:], return_std=True)
    (array([635.5...]), array([0.559...]))
    """

    def __call__(self, xp, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        X : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object
            Left argument of the returned kernel k(X, Y)

        Y : array-like of shape (n_samples_Y, n_features) or list of object,\
            default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            is evaluated instead.

        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the log of
            the kernel hyperparameter is computed.

        Returns
        -------
        K : ndarray of shape (*batch_shape, n_samples_X, n_samples_Y)
            Kernel k(X, Y)

        K_gradient : ndarray of shape (*batch_shape, n_samples_X, n_samples_X, n_dims),\
                optional
            The gradient of the kernel k(X, X) with respect to the log of the
            hyperparameter of the kernel. Only returned when `eval_gradient`
            is True.
        """
        if eval_gradient:
            K, K_gradient = self.kernel(xp, X, Y, eval_gradient=True)
            K_gradient *= self.exponent * K[:, :, np.newaxis] ** (self.exponent - 1)
            return K**self.exponent, K_gradient
        else:
            K = self.kernel(xp, X, Y, eval_gradient=False)
            return K**self.exponent

    def diag(self, xp, X):
        """Returns the diagonal of the kernel k(X, X).

        The result of this method is identical to np.diag(self(X)); however,
        it can be evaluated more efficiently since only the diagonal is
        evaluated.

        Parameters
        ----------
        X : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object
            Argument to the kernel.

        Returns
        -------
        K_diag : ndarray of shape (*batch_shape, n_samples_X,)
            Diagonal of kernel k(X, X)
        """
        return self.kernel.diag(xp, X) ** self.exponent


class ConstantKernel(BatchedKernel, kernels.ConstantKernel):
    """Constant kernel.

    Can be used as part of a product-kernel where it scales the magnitude of
    the other factor (kernel) or as part of a sum-kernel, where it modifies
    the mean of the Gaussian process.

    .. math::
        k(x_1, x_2) = constant\\_value \\;\\forall\\; x_1, x_2

    Adding a constant kernel is equivalent to adding a constant::

            kernel = RBF() + ConstantKernel(constant_value=2)

    is the same as::

            kernel = RBF() + 2


    Read more in the :ref:`User Guide <gp_kernels>`.

    .. versionadded:: 0.18

    Parameters
    ----------
    constant_value : float, default=1.0
        The constant value which defines the covariance:
        k(x_1, x_2) = constant_value

    constant_value_bounds : pair of floats >= 0 or "fixed", default=(1e-5, 1e5)
        The lower and upper bound on `constant_value`.
        If set to "fixed", `constant_value` cannot be changed during
        hyperparameter tuning.

    Examples
    --------
    >>> from sklearn.datasets import make_friedman2
    >>> from sklearn.gaussian_process import GaussianProcessRegressor
    >>> from sklearn.gaussian_process.kernels import RBF, ConstantKernel
    >>> X, y = make_friedman2(n_samples=500, noise=0, random_state=0)
    >>> kernel = RBF() + ConstantKernel(constant_value=2)
    >>> gpr = GaussianProcessRegressor(kernel=kernel, alpha=5,
    ...         random_state=0).fit(X, y)
    >>> gpr.score(X, y)
    0.3696...
    >>> gpr.predict(X[:1,:], return_std=True)
    (array([606.1...]), array([0.24...]))
    """

    def __call__(self, xp, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        X : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object
            Left argument of the returned kernel k(X, Y)

        Y : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object, \
            default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            is evaluated instead.

        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the log of
            the kernel hyperparameter is computed.
            Only supported when Y is None.

        Returns
        -------
        K : ndarray of shape (*batch_shape, n_samples_X, n_samples_Y)
            Kernel k(X, Y)

        K_gradient : ndarray of shape (*batch_shape, n_samples_X, n_samples_X, n_dims), \
            optional
            The gradient of the kernel k(X, X) with respect to the log of the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        """
        if Y is None:
            Y = X
        elif eval_gradient:
            raise ValueError("Gradient can only be evaluated when Y is None.")

        if X.shape[:-2] != Y.shape[:-2]:
            raise ValueError("X and Y must have the same batch dimensions.")
        
        K = xp.full(
            (*X.shape[:-1], Y.shape[-2]),
            self.constant_value,
            dtype=X.dtype
        )
        if eval_gradient:
            if not self.hyperparameter_constant_value.fixed:
                return (
                    K,
                    xp.full(
                        (*X.shape[:-1], X.shape[-2], 1),
                        self.constant_value,
                        dtype=X.dtype
                    ),
                )
            else:
                return K, xp.empty((*X.shape[:-1], X.shape[:-2], 0), dtype=X.dtype)
        else:
            return K

    def diag(self, xp, X):
        """Returns the diagonal of the kernel k(X, X).

        The result of this method is identical to np.diag(self(X)); however,
        it can be evaluated more efficiently since only the diagonal is
        evaluated.

        Parameters
        ----------
        X : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object
            Argument to the kernel.

        Returns
        -------
        K_diag : ndarray of shape (*batch_shape, n_samples_X,)
            Diagonal of kernel k(X, X)
        """
        return xp.full(X.shape[:-1], self.constant_value, dtype=X.dtype)


class WhiteKernel(BatchedKernel, kernels.WhiteKernel):
    """White kernel.

    The main use-case of this kernel is as part of a sum-kernel where it
    explains the noise of the signal as independently and identically
    normally-distributed. The parameter noise_level equals the variance of this
    noise.

    .. math::
        k(x_1, x_2) = noise\\_level \\text{ if } x_i == x_j \\text{ else } 0


    Read more in the :ref:`User Guide <gp_kernels>`.

    .. versionadded:: 0.18

    Parameters
    ----------
    noise_level : float, default=1.0
        Parameter controlling the noise level (variance)

    noise_level_bounds : pair of floats >= 0 or "fixed", default=(1e-5, 1e5)
        The lower and upper bound on 'noise_level'.
        If set to "fixed", 'noise_level' cannot be changed during
        hyperparameter tuning.

    Examples
    --------
    >>> from sklearn.datasets import make_friedman2
    >>> from sklearn.gaussian_process import GaussianProcessRegressor
    >>> from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel
    >>> X, y = make_friedman2(n_samples=500, noise=0, random_state=0)
    >>> kernel = DotProduct() + WhiteKernel(noise_level=0.5)
    >>> gpr = GaussianProcessRegressor(kernel=kernel,
    ...         random_state=0).fit(X, y)
    >>> gpr.score(X, y)
    0.3680...
    >>> gpr.predict(X[:2,:], return_std=True)
    (array([653.0..., 592.1... ]), array([316.6..., 316.6...]))
    """

    def __call__(self, xp, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        X : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object
            Left argument of the returned kernel k(X, Y)

        Y : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object,\
            default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            is evaluated instead.

        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the log of
            the kernel hyperparameter is computed.
            Only supported when Y is None.

        Returns
        -------
        K : ndarray of shape (*batch_shape, n_samples_X, n_samples_Y)
            Kernel k(X, Y)

        K_gradient : ndarray of shape (*batch_shape, n_samples_X, n_samples_X, n_dims),\
            optional
            The gradient of the kernel k(X, X) with respect to the log of the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        """
        if Y is not None and eval_gradient:
            raise ValueError("Gradient can only be evaluated when Y is None.")

        if Y is None:
            K = self.noise_level * xp.tile(xp.eye(X.shape[-2], dtype=X.dtype), (*X.shape[:-2],  1, 1))
            if eval_gradient:
                if not self.hyperparameter_noise_level.fixed:
                    return (
                        K,
                        xp.asarray(K[..., None], copy=True),
                    )
                else:
                    return K, xp.empty((*X.shape[:-1], X.shape[-2], 0), dtype=K.dtype)
            else:
                return K
        else:
            if X.shape[:-2] != Y.shape[:-2]:
                raise ValueError("X and Y must have the same batch dimensions.")
            return xp.zeros((*X.shape[:-1], Y.shape[-2]), dtype=X.dtype)

    def diag(self, xp, X):
        """Returns the diagonal of the kernel k(X, X).

        The result of this method is identical to np.diag(self(X)); however,
        it can be evaluated more efficiently since only the diagonal is
        evaluated.

        Parameters
        ----------
        X : array-like of shape (*batch_shape, n_samples_X, n_features) or list of object
            Argument to the kernel.

        Returns
        -------
        K_diag : ndarray of shape (*batch_shape, n_samples_X,)
            Diagonal of kernel k(X, X)
        """
        return xp.full(X.shape[:-1], self.noise_level, dtype=X.dtype)


class RBF(BatchedKernel, kernels.RBF):
    """Radial basis function kernel (aka squared-exponential kernel).

    The RBF kernel is a stationary kernel. It is also known as the
    "squared exponential" kernel. It is parameterized by a length scale
    parameter :math:`l>0`, which can either be a scalar (isotropic variant
    of the kernel) or a vector with the same number of dimensions as the inputs
    X (anisotropic variant of the kernel). The kernel is given by:

    .. math::
        k(x_i, x_j) = \\exp\\left(- \\frac{d(x_i, x_j)^2}{2l^2} \\right)

    where :math:`l` is the length scale of the kernel and
    :math:`d(\\cdot,\\cdot)` is the Euclidean distance.
    For advice on how to set the length scale parameter, see e.g. [1]_.

    This kernel is infinitely differentiable, which implies that GPs with this
    kernel as covariance function have mean square derivatives of all orders,
    and are thus very smooth.
    See [2]_, Chapter 4, Section 4.2, for further details of the RBF kernel.

    Read more in the :ref:`User Guide <gp_kernels>`.

    .. versionadded:: 0.18

    Parameters
    ----------
    length_scale : float or ndarray of shape (n_features,), default=1.0
        The length scale of the kernel. If a float, an isotropic kernel is
        used. If an array, an anisotropic kernel is used where each dimension
        of l defines the length-scale of the respective feature dimension.

    length_scale_bounds : pair of floats >= 0 or "fixed", default=(1e-5, 1e5)
        The lower and upper bound on 'length_scale'.
        If set to "fixed", 'length_scale' cannot be changed during
        hyperparameter tuning.

    References
    ----------
    .. [1] `David Duvenaud (2014). "The Kernel Cookbook:
        Advice on Covariance functions".
        <https://www.cs.toronto.edu/~duvenaud/cookbook/>`_

    .. [2] `Carl Edward Rasmussen, Christopher K. I. Williams (2006).
        "Gaussian Processes for Machine Learning". The MIT Press.
        <http://www.gaussianprocess.org/gpml/>`_

    Examples
    --------
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.gaussian_process import GaussianProcessClassifier
    >>> from sklearn.gaussian_process.kernels import RBF
    >>> X, y = load_iris(return_X_y=True)
    >>> kernel = 1.0 * RBF(1.0)
    >>> gpc = GaussianProcessClassifier(kernel=kernel,
    ...         random_state=0).fit(X, y)
    >>> gpc.score(X, y)
    0.9866...
    >>> gpc.predict_proba(X[:2,:])
    array([[0.8354..., 0.03228..., 0.1322...],
           [0.7906..., 0.0652..., 0.1441...]])
    """
    
    def __call__(self, xp, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        X : ndarray of shape (*batch_shape, n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)

        Y : ndarray of shape (n_samples_Y, n_features), default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.

        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the log of
            the kernel hyperparameter is computed.
            Only supported when Y is None.

        Returns
        -------
        K : ndarray of shape (*batch_shape, n_samples_X, n_samples_Y)
            Kernel k(X, Y)

        K_gradient : ndarray of shape (*batch_shape, n_samples_X, n_samples_X, n_dims), \
                optional
            The gradient of the kernel k(X, X) with respect to the log of the
            hyperparameter of the kernel. Only returned when `eval_gradient`
            is True.
        """
        # X = xp.atleast_2d(X)
        length_scale = xp.asarray(_check_length_scale(X, self.length_scale), dtype=X.dtype)
        X = X / length_scale
        if Y is None:
            if not eval_gradient or not self.anisotropic or length_scale.shape[0] == 1:
                # In this case we don't need the full distance matrix, we can reduce over the features dimension
                dists = xp.cdist(X, X) ** 2
                K = xp.exp(-0.5 * dists)
            else:
                dists = ((X[..., None, :]) - X[..., None, :, :]) ** 2
                K = xp.exp(-0.5 * dists.sum(-1))
        else:
            if eval_gradient:
                raise ValueError("Gradient can only be evaluated when Y is None.")
            Y = Y / length_scale
            dists = xp.cdist(X, Y) ** 2
            K = xp.exp(-0.5 * dists)

        if eval_gradient:
            if self.hyperparameter_length_scale.fixed:
                # Hyperparameter l kept fixed
                return K, xp.empty((X.shape[0], X.shape[0], 0), dtype=K.dtype)
            elif not self.anisotropic or length_scale.shape[0] == 1:
                K_gradient = (K * dists)[..., None]
                return K, K_gradient
            elif self.anisotropic:
                K_gradient = K[..., None] * dists
                return K, K_gradient
        else:
            return K
        
    def diag(self, xp, X):
        """Returns the diagonal of the kernel k(X, X).

        The result of this method is identical to np.diag(self(X)); however,
        it can be evaluated more efficiently since only the diagonal is
        evaluated.

        Parameters
        ----------
        X : ndarray of shape (*batch_shape, n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)

        Returns
        -------
        K_diag : ndarray of shape (*batch_shape, n_samples_X,)
            Diagonal of kernel k(X, X)
        """
        return xp.ones(X.shape[:-1], dtype=X.dtype)
        

class Matern(RBF, kernels.Matern):
    """Matern kernel.

    The class of Matern kernels is a generalization of the :class:`RBF`.
    It has an additional parameter :math:`\\nu` which controls the
    smoothness of the resulting function. The smaller :math:`\\nu`,
    the less smooth the approximated function is.
    As :math:`\\nu\\rightarrow\\infty`, the kernel becomes equivalent to
    the :class:`RBF` kernel. When :math:`\\nu = 1/2`, the Matérn kernel
    becomes identical to the absolute exponential kernel.
    Important intermediate values are
    :math:`\\nu=1.5` (once differentiable functions)
    and :math:`\\nu=2.5` (twice differentiable functions).

    The kernel is given by:

    .. math::
         k(x_i, x_j) =  \\frac{1}{\\Gamma(\\nu)2^{\\nu-1}}\\Bigg(
         \\frac{\\sqrt{2\\nu}}{l} d(x_i , x_j )
         \\Bigg)^\\nu K_\\nu\\Bigg(
         \\frac{\\sqrt{2\\nu}}{l} d(x_i , x_j )\\Bigg)



    where :math:`d(\\cdot,\\cdot)` is the Euclidean distance,
    :math:`K_{\\nu}(\\cdot)` is a modified Bessel function and
    :math:`\\Gamma(\\cdot)` is the gamma function.
    See [1]_, Chapter 4, Section 4.2, for details regarding the different
    variants of the Matern kernel.

    Read more in the :ref:`User Guide <gp_kernels>`.

    .. versionadded:: 0.18

    Parameters
    ----------
    length_scale : float or ndarray of shape (n_features,), default=1.0
        The length scale of the kernel. If a float, an isotropic kernel is
        used. If an array, an anisotropic kernel is used where each dimension
        of l defines the length-scale of the respective feature dimension.

    length_scale_bounds : pair of floats >= 0 or "fixed", default=(1e-5, 1e5)
        The lower and upper bound on 'length_scale'.
        If set to "fixed", 'length_scale' cannot be changed during
        hyperparameter tuning.

    nu : float, default=1.5
        The parameter nu controlling the smoothness of the learned function.
        The smaller nu, the less smooth the approximated function is.
        For nu=inf, the kernel becomes equivalent to the RBF kernel and for
        nu=0.5 to the absolute exponential kernel. Important intermediate
        values are nu=1.5 (once differentiable functions) and nu=2.5
        (twice differentiable functions). Note that values of nu not in
        [0.5, 1.5, 2.5, inf] incur a considerably higher computational cost
        (appr. 10 times higher) since they require to evaluate the modified
        Bessel function. Furthermore, in contrast to l, nu is kept fixed to
        its initial value and not optimized.

    References
    ----------
    .. [1] `Carl Edward Rasmussen, Christopher K. I. Williams (2006).
        "Gaussian Processes for Machine Learning". The MIT Press.
        <http://www.gaussianprocess.org/gpml/>`_

    Examples
    --------
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.gaussian_process import GaussianProcessClassifier
    >>> from sklearn.gaussian_process.kernels import Matern
    >>> X, y = load_iris(return_X_y=True)
    >>> kernel = 1.0 * Matern(length_scale=1.0, nu=1.5)
    >>> gpc = GaussianProcessClassifier(kernel=kernel,
    ...         random_state=0).fit(X, y)
    >>> gpc.score(X, y)
    0.9866...
    >>> gpc.predict_proba(X[:2,:])
    array([[0.8513..., 0.0368..., 0.1117...],
            [0.8086..., 0.0693..., 0.1220...]])
    """

    def __call__(self, xp, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        X : ndarray of shape (*batch_shape, n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)

        Y : ndarray of shape (n_samples_Y, n_features), default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.

        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the log of
            the kernel hyperparameter is computed.
            Only supported when Y is None.

        Returns
        -------
        K : ndarray of shape (*batch_shape, n_samples_X, n_samples_Y)
            Kernel k(X, Y)

        K_gradient : ndarray of shape (*batch_shape, n_samples_X, n_samples_X, n_dims), \
                optional
            The gradient of the kernel k(X, X) with respect to the log of the
            hyperparameter of the kernel. Only returned when `eval_gradient`
            is True.
        """
        
        # X = xp.atleast_2d(X)
        length_scale = xp.asarray(kernels._check_length_scale(X, self.length_scale), dtype=X.dtype)
        full_dists = None
        should_make_squareform = False
        if Y is None:
            # dists = xp.nn.functional.pdist(X / length_scale, p=2)
            if eval_gradient and self.anisotropic:
                # Full distances will be needed afterwards
                full_dists = (X[..., None, :] - X[..., None, :, :]) ** 2 / (length_scale**2)

                def get_dists():
                    return xp.sqrt(full_dists.sum(-1))
                
                def get_sqdists():
                    return full_dists.sum(-1)
                
            else:
                tmp = X / length_scale
                dists = xp.cdist(tmp, tmp)
                del tmp

                def get_dists():
                    return dists
                
                def get_sqdists():
                    return dists ** 2
            
        else:
            if eval_gradient:
                raise ValueError("Gradient can only be evaluated when Y is None.")
            dists = xp.cdist(X / length_scale, Y / length_scale)

            def get_dists():
                return dists
            
            def get_sqdists():
                return dists ** 2

        if self.nu == 0.5:
            K = xp.exp(-get_dists())
        elif self.nu == 1.5:
            K = get_dists() * math.sqrt(3)
            K = (1.0 + K) * xp.exp(-K)
        elif self.nu == 2.5:
            K = get_dists() * math.sqrt(5)
            K = (1.0 + K + K**2 / 3.0) * xp.exp(-K)
        elif self.nu == np.inf:
            K = xp.exp(-get_sqdists() / 2.0)
        else:  # general case; expensive to evaluate
            raise NotImplementedError()
            # K = dists
            # K[K == 0.0] += np.finfo(float).eps  # strict zeros result in nan
            # tmp = math.sqrt(2 * self.nu) * K
            # K.fill((2 ** (1.0 - self.nu)) / gamma(self.nu))
            # K *= tmp**self.nu
            # K *= kv(self.nu, tmp)

        if should_make_squareform:
            K = _squareform(xp, K, X.shape[-2])

        if eval_gradient:
            if self.hyperparameter_length_scale.fixed:
                # Hyperparameter l kept fixed
                K_gradient = xp.empty((X.shape[0], X.shape[0], 0), dtype=K.dtype)
                return K, K_gradient

            # We need to recompute the pairwise dimension-wise distances
            if self.anisotropic:
                D = full_dists
            else:
                # D = squareform(dists**2)[:, :, None]
                D = get_sqdists()[..., None]

            if self.nu == 0.5:
                denominator = xp.sqrt(D.sum(axis=-1))
                denominator[denominator == 0] = 1  # Prevent division by zero
                K_gradient = K[..., None] * (D / denominator[..., None])
                del denominator
            elif self.nu == 1.5:
                K_gradient = 3 * D * xp.exp(-xp.sqrt(3 * D.sum(-1)))[..., None]
            elif self.nu == 2.5:
                tmp = xp.sqrt(5 * D.sum(-1))[..., None]
                K_gradient = 5.0 / 3.0 * D * (tmp + 1) * xp.exp(-tmp)
                del tmp
            elif self.nu == np.inf:
                K_gradient = D * K[..., None]
            else:
                raise NotImplementedError()
                # approximate gradient numerically
                # def f(theta):  # helper function
                #     return self.clone_with_theta(theta)(X, Y)

                # return K, _approx_fprime(self.theta, f, 1e-10)

            if not self.anisotropic:
                return K, K_gradient.sum(-1)[..., None]
            else:
                return K, K_gradient
        else:
            return K


# TODO: implement remainign kernels, mainly RationalQuadratic, ExpSine, DotProduct, PairWiseK


def _squareform(xp, x, n):
    """Convert a 1D array to a square matrix."""
    out = xp.empty((*x.shape[:-1], n, n), dtype=x.dtype)
    ind = np.tril_indices(n, k=-1)
    assert ind[0].size * np.cumprod(out.shape[:-2]) == x.size, "Unexpected error. Contact the developer"
    out[..., ind[0], ind[1]] = x
    del ind
    out += out.transpose(-2, -1)
    return out


def _check_length_scale(X, length_scale):
    length_scale = np.squeeze(length_scale).astype(float)
    if np.ndim(length_scale) > 1:
        raise ValueError("length_scale cannot be of dimension greater than 1")
    if np.ndim(length_scale) == 1 and X.shape[-1] != length_scale.shape[0]:
        raise ValueError(
            "Anisotropic kernel must have the same number of "
            "dimensions as data (%d!=%d)" % (length_scale.shape[0], X.shape[1])
        )
    return length_scale



""" This module implements specific operations on batched kernels that are optmized for Cuda hardware through
custom Cuda code. Specifically, this module provides optimized functions for nearest neighbour search and kernel 
evaluations.
"""

import numbers
import copy

import numpy as np
from sklearn.gaussian_process import kernels

try:
    import cupy as cp
except ImportError:
    raise ImportError("This module requires cupy. Install with `pip install nngpr[cupy]`")


class CudaKernelAccelerator:
    """ Class that provides the end-user interface to NngprWorker instances. This class collects the different 
    kernel optimization classes (currently only RBF is available) and dispatches the accelerated functions to the right
    kernel class
    """

    def __init__(self):
        super().__init__()
        self._last_kernel = None
        self._last_theta = np.nan
        self._last_params = None
        self._last_mapping = None
        self._acc_fn = None

    def can_accelerate_kernel(self, kernel):
        """ Indicates whether accelerated functions are available for the given kernel 
        
        Parameters
        ----------
        kernel : nngpr.base_nngpr.NngprWorker
            Gaussian process kernel

        Returns
        -------

        bool : True if accelerated functions are available for the kernel in input
        
        """
        self._update_kernel_if_needed(kernel)
        return self._acc_fn is not None

    def get_kernel_params(self, kernel):
        """ Helper functions that extract the kernel parameters and places them in a dictionary. 

        Parameters
        ----------
        kernel : nngpr.base_nngpr.NngprWorker
            Input kernel whose parameters will be extracted and returned as output
        
        Returns
        -------
        kernel_parameters : dict
            Dictionary that maps each kernel parameter (name) to its value, e.g.:
            {'noise_level': 1.234, 'length_scale': [1.8,3.3,6.2,8.3]}

        """
        self._update_kernel_if_needed(kernel)
        return self._last_params

    def get_kernel_theta_mapping(self, kernel):
        """ Helper functions that maps each parameter retruned by 'get_kernel_params' to the corresponding value 
        in kernel.theta

        Parameters
        ----------
        kernel : nngpr.base_nngpr.NngprWorker
            Input kernel whose parameters will be mapped to kernel.theta
        
        Returns
        -------
        kernel_parameters : dict
            Dictionary that maps each kernel parameter (name) to its index in kernel.theta, e.g.
            {'noise_level': 0, 'length_scale': [1,2,3,4]}

        """
        self._update_kernel_if_needed(kernel)
        return self._last_mapping
    
    def find_nearest_neighbors(self, cuda_device_index, ref, query, nn_type, kernel, num_nn, ref_cut_index):
        """ Implements nngpr.base_nngpr.WorkerNngpr.find_nearest_neighbors

        Parameters
        ----------

        cuda_device_index : int
            Cuda device to use for calculations
        
        ref : any array type implementing the dlpack interface.
            see nngpr.base_nngpr.WorkerNngpr docstring.

        query : any array type implementing the dlpack interface.
            see nngpr.base_nngpr.WorkerNngpr docstring

        nn_type : see nngpr.base_nngpr.WorkerNngpr docstring

        kernel : see nngpr.base_nngpr.WorkerNngpr docstring

        num_nn : see nngpr.base_nngpr.WorkerNngpr docstring

        ref_cut_index : see nngpr.base_nngpr.WorkerNngpr docstring

        Returns
        -------
        Same as nngpr.base_nngpr.WorkerNngpr.find_nearest_neighbors. Raises NotImplementedError if the kernel cannot be accelerated.
        
        """
        if not self.can_accelerate_kernel(kernel):
            raise NotImplementedError("Cannot accelerate this kernel")
        return self._acc_fn[0](cuda_device_index, ref, query, nn_type, kernel, num_nn, ref_cut_index)
    
    def fill_nn_kernel(self, cuda_device_index, x, nn_indices, kernel, eval_gradient=False):
        """ Implements nngpr.base_nngpr.WorkerNngpr.fill_nn_kernel

        Parameters
        ----------

        cuda_device_index : int
            Cuda device to use for calculations
        
        x : any array type implementing the dlpack interface.
            see nngpr.base_nngpr.WorkerNngpr docstring

        nn_indices : any array type implementing the dlpack interface.
            see nngpr.base_nngpr.WorkerNngpr docstring

        kernel : see nngpr.base_nngpr.WorkerNngpr docstring

        eval_gradient : see nngpr.base_nngpr.WorkerNngpr docstring

        Returns
        -------
        Same as nngpr.base_nngpr.WorkerNngpr.fill_nn_kernel. Raises NotImplementedError if the kernel cannot be accelerated.
        
        """
        if not self.can_accelerate_kernel(kernel):
            raise NotImplementedError("Cannot accelerate this kernel")
        return self._acc_fn[1](cuda_device_index, x, nn_indices, kernel, eval_gradient=eval_gradient)
    
    def _update_kernel_if_needed(self, kernel):
        """ Extracts the kernel parameters and theta mapping for this kernel and stores it in instance attributes """
        if self._last_kernel is kernel and np.all(kernel.theta == self._last_theta):
            return
        self._acc_fn = None
        self._last_kernel = kernel
        self._last_theta = copy.deepcopy(self._last_kernel.theta)
        self._last_params = None
        self._last_mapping = None

        try:
            params = RbfAccelerator.get_kernel_params(kernel)
            theta_mapping = RbfAccelerator.get_kernel_theta_mapping(kernel)
            self._last_params = params
            self._last_mapping = theta_mapping
            self._acc_fn = (RbfAccelerator.find_nearest_neighbors, RbfAccelerator.fill_nn_kernel)

        except NotImplementedError:
            pass


class RbfAccelerator:
    """ This class implements Cuda accelerated functions for kernles that can be casted to the form:
        WhiteKernel + ConstantKernel * RBF
    """

    @staticmethod
    def can_accelerate_kernel(kernel):
        """ Indicates whether accelerated functions are available for the given kernel. Only kernels
        of the form WhiteKernel + ConstantKernel * RBF can be accelerated by this class
        
        Parameters
        ----------
        kernel : nngpr.base_nngpr.NngprWorker
            Gaussian process kernel

        Returns
        -------

        bool : True if accelerated functions are available for the kernel in input
        
        """
        try:
            RbfAccelerator.get_kernel_params(kernel)
            RbfAccelerator.get_kernel_theta_mapping(kernel)
        except NotImplementedError:
            return False
        
        return True
        
    @staticmethod
    def find_nearest_neighbors(cuda_device_index, ref, query, nn_type, kernel, num_nn, ref_cut_index):
        """ Implements nngpr.base_nngpr.WorkerNngpr.find_nearest_neighbors

        Parameters
        ----------

        cuda_device_index : int
            Cuda device to use for calculations
        
        ref : any array type implementing the dlpack interface.
            see nngpr.base_nngpr.WorkerNngpr docstring.

        query : any array type implementing the dlpack interface.
            see nngpr.base_nngpr.WorkerNngpr docstring

        nn_type : see nngpr.base_nngpr.WorkerNngpr docstring

        kernel : see nngpr.base_nngpr.WorkerNngpr docstring

        num_nn : see nngpr.base_nngpr.WorkerNngpr docstring

        ref_cut_index : see nngpr.base_nngpr.WorkerNngpr docstring

        Returns
        -------
        Same as nngpr.base_nngpr.WorkerNngpr.find_nearest_neighbors. Raises NotImplementedError if the kernel cannot be accelerated.
        
        """

        kernel_params = RbfAccelerator.get_kernel_params(kernel)

        with cp.cuda.Device(cuda_device_index):

            # Check if input is cupy array
            ref, query = _ensure_cupy_array(ref), _ensure_cupy_array(query)
            
            data_type = np.promote_types(ref.dtype, query.dtype)

            if nn_type == 'kernel-space' and 'length_scale' in kernel_params:
                ls = kernel_params['length_scale']
                if type(ls) is np.ndarray:
                    ls = cp.array(ls, dtype=data_type)
                ref, query = ref / ls, query / ls

            # Start building the kernel header
            defines = {} 
            if data_type == np.dtype(np.float64):
                defines['DATA_TYPE'] = "double"
                defines['INFINITY'] = '__longlong_as_double(0x7ff0000000000000ULL)'
                defines['NEG_INFINITY'] = '__longlong_as_double(0xfff0000000000000ULL)'
            elif data_type == np.dtype(np.float32):
                defines['DATA_TYPE'] = "float"
                defines['INFINITY'] = '__int_as_float(0x7f800000)'
                defines['NEG_INFINITY'] = '__int_as_float(0xff800000)'
            elif data_type == np.dtype(np.float16):
                defines['DATA_TYPE'] = "half"
                defines['INFINITY'] = '__short_as_half(0x7c00)'
                defines['NEG_INFINITY'] = '__short_as_half(0xfc00)'
            if ref.itemsize > query.itemsize:
                query = query.astype(ref.dtype)
            if ref.itemsize < query.itemsize:
                ref = ref.astype(query.dtype)

            # Add float16 support
            kernel_headers = []
            if ref.itemsize == 2:
                kernel_headers.append('#include <cuda_fp16.h>')

            # Helper function
            def _get_dtype_for_range(x):
                if x <= np.iinfo(np.int32).max:
                    cuda_type = "int"
                    cupy_type = cp.intc
                elif x <= np.iinfo(np.int64).max:
                    # signed integers somehow performs better than unsigned
                    cuda_type = "long long"
                    cupy_type = cp.int64
                else:
                    assert x <= np.iinfo(np.uint64).max, "Unhandled case of size_t overflow in cuda kernel"
                    cuda_type = "size_t"
                    cupy_type = cp.intp

                return cuda_type, cupy_type

            # Define types
            max_int_32 = np.iinfo(np.int32).max
            assert ref.shape[1] < max_int_32, "Unhandled case of int overflow in cuda kernel"
            assert num_nn < max_int_32, "Unhandled case of int overflow in cuda kernel"
            defines['INDEX_TYPE'], index_type = _get_dtype_for_range(max(query.shape[0], ref.shape[0]))
            defines['N_QUERY'] = query.shape[0]
            defines['N_REF'] = ref.shape[0]
            defines['NUM_K'] = num_nn
            defines['NUM_FEATURE'] = ref.shape[1]

            # Create common variables
            query_nb = query.shape[0]
            index_gpu = cp.empty(query.shape[0] * num_nn, dtype=index_type).reshape((num_nn, query.shape[0]))
            use_x_transposed = True

            # Number of NNs to keep in cache
            k_for_cache = num_nn

            block_dim_nn = 256
            shared_mem_req_size = k_for_cache * block_dim_nn * (data_type.itemsize + index_type(1).itemsize)
            defines['SIZE_TYPE'], _ = _get_dtype_for_range(max(query.size, ref.size, k_for_cache * block_dim_nn))

            block_size = (block_dim_nn,)
            grid_size = (int(np.ceil(query_nb / block_size[0])),)

            opt_call_kwargs = {}
            opt_call_k_args = []
            cache_dist, cache_index = None, None

            # TODO: get the right static memory size for the device
            max_static_mem_size: int = 49152
            max_static_mem_size_optin: int = 98304
            
            if shared_mem_req_size <= max_static_mem_size:
                cache_memory_type = 'shared_static'
            elif shared_mem_req_size <= max_static_mem_size_optin * 0.9:
                cache_memory_type = 'shared_dynamic'
            else:
                cache_memory_type = 'global'

            if cache_memory_type == 'shared_static':
                defines['USE_STATIC_SHARED_MEM_CACHE'] = 'true'
                dynamic_shared_mem_size = 0
            elif cache_memory_type == 'shared_dynamic':
                defines['USE_DYNAMIC_SHARED_MEM_CACHE'] = 'true'
                dynamic_shared_mem_size = shared_mem_req_size
                opt_call_kwargs = {'shared_mem': dynamic_shared_mem_size}
            elif cache_memory_type == 'global':
                cache_dist = cp.empty(shape=(k_for_cache, query_nb), dtype=data_type)
                cache_index = cp.empty(shape=(k_for_cache, query_nb), dtype=index_type)
                cache_dist[:] = 0
                defines.update({
                    'USE_GLOBAL_MEM_CACHE': 'true',
                    'CACHE_DIST_PITCH': '*' + str(cache_dist.shape[1]), 
                    'CACHE_INDEX_PITCH': '*' + str(cache_index.shape[1])
                })
                dynamic_shared_mem_size = 0
                opt_call_k_args = [cache_dist, cache_index]
            else:
                raise RuntimeError("Unexpcted cache_memory_type: " + str(cache_memory_type))

            if use_x_transposed:
                defines['REF_PITCH_FEATURE'] = '* ' + str(ref.shape[0])  # ideally query should be in a pitched memory layout
                defines['REF_PITCH_POINT'] = ''
            else:
                defines['REF_PITCH_FEATURE'] = ''
                defines['REF_PITCH_POINT'] = '* ' + str(ref.shape[0])  # ideally query should be in a pitched memory layout

            # Initialize veriables
            if use_x_transposed:
                ref = ref.T
            ref = cp.array(cp.ascontiguousarray(ref) if type(ref) is cp.ndarray else np.ascontiguousarray(ref))
            query = query.T
            query = cp.array(cp.ascontiguousarray(query) if type(query) is cp.ndarray else np.ascontiguousarray(query))
            query_pitch = query.shape[1]  # ideally query should be in a pitched memory layout

            # Determine max search point
            if ref_cut_index is None:
                defines['MAX_SEARCH_POINT'] = defines['N_REF']
            else:
                defines['MAX_SEARCH_POINT'] = '(i_query + ' + str(ref_cut_index) + ')'
                index_gpu[:, :max(0, num_nn - ref_cut_index)] = -1

            # Build the kernel with the desired options
            defines['BLOCK_DIM'] = block_dim_nn
            defines['QUERY_PITCH'] = query_pitch
            defines['NUM_K_FOR_CACHE'] = k_for_cache
            defines['BLOCK_DIM'] = block_dim_nn
            defines['INDEX_PITCH'] = index_gpu.shape[1]

            kernel_headers = kernel_headers + ["#define " + k + " " + str(v) for k, v in defines.items()] + ["", ""]
            kernel_headers = "\n".join(kernel_headers)
            krn = cp.RawKernel(kernel_headers + _calc_nn_rbf_kernel_text, 'calc_nn')

            # Run
            if dynamic_shared_mem_size > 0:
                krn.max_dynamic_shared_size_bytes = dynamic_shared_mem_size
            krn(
                grid_size, block_size, 
                (ref, query, index_gpu, *opt_call_k_args), **opt_call_kwargs)
            return cp.transpose(index_gpu)

    @staticmethod
    def fill_nn_kernel(cuda_device_index, x, nn_indices, kernel, eval_gradient=False):
        """ Implements nngpr.base_nngpr.WorkerNngpr.fill_nn_kernel

        Parameters
        ----------

        cuda_device_index : int
            Cuda device to use for calculations
        
        x : any array type implementing the dlpack interface.
            see nngpr.base_nngpr.WorkerNngpr docstring

        nn_indices : any array type implementing the dlpack interface.
            see nngpr.base_nngpr.WorkerNngpr docstring

        kernel : see nngpr.base_nngpr.WorkerNngpr docstring

        eval_gradient : see nngpr.base_nngpr.WorkerNngpr docstring

        Returns
        -------
        Same as nngpr.base_nngpr.WorkerNngpr.fill_nn_kernel. Raises NotImplementedError if the kernel cannot be accelerated.
        
        """

        with cp.cuda.Device(cuda_device_index):

            kernel_params = RbfAccelerator.get_kernel_params(kernel)
            kernel_theta_mapping = RbfAccelerator.get_kernel_theta_mapping(kernel)
            nn_indices = nn_indices.T  # For faster cuda access

            # Check if input is cupy array
            x, nn_indices = _ensure_cupy_array(x), _ensure_cupy_array(nn_indices)
            
            # Prepare data to be accessed by the kernel
            if not x.flags['C_CONTIGUOUS']:
                x = cp.ascontiguousarray(x)
            assert x.flags['C_CONTIGUOUS']
            if not nn_indices.flags['C_CONTIGUOUS']:
                nn_indices = cp.ascontiguousarray(nn_indices)
            assert nn_indices.flags['C_CONTIGUOUS']
            assert nn_indices.dtype in [np.int32, np.int64], "Unsupported dtype for nn_indices: " + str(nn_indices.dtype)

            # Allocate output
            n_theta = sum(v.size if type(v) is np.ndarray else 1 for k, v in kernel_theta_mapping.items())
            nn = nn_indices.shape[0]
            nn_kernel = cp.zeros((nn, nn, nn_indices.shape[1]), dtype=x.dtype)
            nn_kernel_grad = cp.zeros((nn, nn, n_theta, nn_indices.shape[1]), dtype=x.dtype) if eval_gradient else cp.zeros((0, 0, 0, 0))
            rbf_length_is_vec = not isinstance(kernel_params.get('length_scale', 1), numbers.Number)
            rbf_length_scales = cp.array(kernel_params['length_scale'].reshape(-1), dtype=x.dtype) if rbf_length_is_vec \
                else cp.array(kernel_params.get('length_scale', 1), dtype=x.dtype)

            # Check dtypes
            if x.dtype.type == np.float64:
                data_type = 'double'
                exp_fun = 'exp'
            elif x.dtype.type == np.float32:
                data_type = 'float'
                exp_fun = 'expf'
            elif x.dtype.type == np.float16:
                data_type = 'single'
                exp_fun = 'h2exp'
            else:
                raise NotImplementedError(str(x.dtype))

            min_i_type = np.min_scalar_type(max(x.shape[0], x.shape[1], nn_indices.shape[0], nn_indices.shape[1]))
            if np.can_cast(min_i_type, 'int32'):
                index_type = 'int'
            elif np.can_cast(min_i_type, 'int64'):
                index_type = 'long long'
            else:
                raise NotImplementedError(str(min_i_type))

            min_i_type = np.min_scalar_type(max(nn_kernel.size, nn_kernel_grad.size))
            if np.can_cast(min_i_type, 'int32'):
                big_index_type = 'int'
            elif np.can_cast(min_i_type, 'int64'):
                big_index_type = 'long long'
            elif np.can_cast(min_i_type, 'uint64'):
                big_index_type = 'size_t'
            else:
                raise NotImplementedError(str(min_i_type))

            # Build kernel text
            block_dim_nn = 64
            defines = {
                'BLOCK_DIM': block_dim_nn,
                'EXP': exp_fun,

                'NUM_REF_POINTS': nn_indices.shape[1],
                'NUM_NN': nn_indices.shape[0],
                'NUM_FEATURES': x.shape[1],
                'NUM_THETA': n_theta,

                'PITCH_NN_POINTS': nn_indices.shape[1],
                'PITCH_DATA_POINTS': x.shape[1],
                'PITCH_NN_KERNEL': nn_kernel.shape[-1],
                'PITCH_NN_KERNEL_GRAD': nn_kernel_grad.shape[-1] * nn_kernel_grad.shape[-2],
                'THETA_PITCH': nn_kernel_grad.shape[-1],

                'DATA_TYPE': data_type,
                'INDEX_TYPE': index_type,
                'NN_IND_DTYPE': 'long long' if nn_indices.dtype == np.int64 else 'int',
                'BIG_INDEX_TYPE': big_index_type,

                'WHITE_NOISE_INDEX': kernel_theta_mapping.get('noise_level', -1) if eval_gradient else -1,
                'CONSTANT_LEVEL_INDEX': kernel_theta_mapping.get('constant_value', -1) if eval_gradient else -1,
                'RBF_LEVEL_INDEX': kernel_theta_mapping.get('rbf_level', -1) if eval_gradient else -1,
                'RBF_SCALE_INDEX': kernel_theta_mapping.get('length_scale', -1) if eval_gradient and not rbf_length_is_vec else -1,
                'RBF_SCALE_VEC_INDEX_0': kernel_theta_mapping.get('length_scale', [-1])[0] if eval_gradient and rbf_length_is_vec else -1,
                'RBF_SCALE_VEC_INDEX_1': kernel_theta_mapping.get('length_scale', [-1])[-1] + 1 if eval_gradient and rbf_length_is_vec else -1,
            }
            if data_type == np.float16:
                defines['USE_F16'] = 1
            if rbf_length_is_vec:
                defines['IS_VEC_RBF_SCALE'] = 1
            lines = ["#define " + k + " " + str(v) for k, v in defines.items()] + ["", ""]
            kernel_text = "\n".join(lines) + _fill_nn_RBF_kernel_text

            # Execute kernel
            krn = cp.RawKernel(kernel_text, 'fill_nn_kernel')
            block_size = (block_dim_nn,)
            grid_size = (int(np.ceil(nn_indices.shape[1] / block_size[0])),)

            krn(grid_size, block_size, (
                x, nn_indices, rbf_length_scales, nn_kernel, nn_kernel_grad,
                x.dtype.type(kernel_params.get('constant_value', 0)),
                x.dtype.type(kernel_params.get('noise_level', 0)),
                x.dtype.type(kernel_params.get('rbf_level', 1 if 'length_scale' in kernel_params else 0))))

            # Post-processing
            nn_kernel = cp.moveaxis(nn_kernel, -1, 0)  # Follow sickitlearn convention
            nn_kernel_grad = cp.moveaxis(nn_kernel_grad, -1, 0)  # Follow sickitlearn convention

            return nn_kernel, nn_kernel_grad

    @staticmethod
    def get_kernel_params(kernel):
        """ Helper functions that extract the kernel parameters and places them in a dictionary. 

        Parameters
        ----------
        kernel : nngpr.base_nngpr.NngprWorker
            Input kernel whose parameters will be extracted and returned as output
        
        Returns
        -------
        kernel_parameters : dict
            Dictionary that maps each kernel parameter (name) to its value, e.g.:
            {'noise_level': 1.234, 'length_scale': [1.8,3.3,6.2,8.3]}

        """

        if issubclass(type(kernel), kernels.WhiteKernel):
            kernel: kernels.WhiteKernel
            return {'noise_level': kernel.noise_level}

        if issubclass(type(kernel), kernels.ConstantKernel):
            kernel: kernels.ConstantKernel
            return {'constant_value': kernel.constant_value}

        if issubclass(type(kernel), kernels.RBF):
            kernel: kernels.RBF
            return {'length_scale': kernel.length_scale if isinstance(kernel.length_scale, numbers.Number)
            else np.array(kernel.length_scale).reshape((1, -1))}

        if issubclass(type(kernel), kernels.Sum):
            kernel: kernels.Sum
            p1 = RbfAccelerator.get_kernel_params(kernel.k1)
            p2 = RbfAccelerator.get_kernel_params(kernel.k2)
            # Sum is supported only with a white kernel
            is_p1_rbf = len(set(p1.keys()).difference({'noise_level', 'constant_value'})) > 0
            is_p2_rbf = len(set(p2.keys()).difference({'noise_level', 'constant_value'})) > 0
            if is_p1_rbf and is_p2_rbf:
                raise NotImplementedError("Sum of kernels is not supported between RBF kernels. Got sum between "
                                   + str(p1) + " and " + str(p2))
            if is_p1_rbf:
                p1, p2 = p2, p1
            # p1 is the non rbf kernel
            if ('noise_level' in p1) or ('noise_level' in p2):
                p2['noise_level'] = p1.get('noise_level', 0) + p2.get('noise_level', 0)
            if ('constant_value' in p1) or ('constant_value' in p2):
                p2['constant_value'] = p1.get('constant_value', 0) + p2.get('constant_value', 0)
            return p2

        if issubclass(type(kernel), kernels.Product):
            kernel: kernels.Product
            p1 = RbfAccelerator.get_kernel_params(kernel.k1)
            p2 = RbfAccelerator.get_kernel_params(kernel.k2)
            # Product is supported only with a ConstantKernel
            is_p1_const = len(set(p1.keys()).difference({'constant_value'})) == 0
            is_p2_const = len(set(p2.keys()).difference({'constant_value'})) == 0
            if not is_p1_const and not is_p2_const:
                raise NotImplementedError("Product of kernels is supported only with a ConstantKernel. Got product between "
                                   + str(p1) + " and " + str(p2))
            if is_p2_const:
                p1, p2 = p2, p1
            # p1 is the constant kernel
            if 'noise_level' in p2:
                p2['noise_level'] = p1['constant_value'] * p2['noise_level']
            if 'constant_value' in p2:
                p2['constant_value'] = p1['constant_value'] * p2['constant_value']
            if ('rbf_level' in p2) or ('length_scale' in p2):
                p2['rbf_level'] = p1['constant_value'] * p2.get('rbf_level', 1)
            return p2

        raise NotImplementedError("Unsupported kernel type: " + str(type(kernel)))
    
    @staticmethod
    def get_kernel_theta_mapping(kernel):
        """ Helper functions that maps each parameter retruned by 'get_kernel_params' to the corresponding value 
        in kernel.theta

        Parameters
        ----------
        kernel : nngpr.base_nngpr.NngprWorker
            Input kernel whose parameters will be mapped to kernel.theta
        
        Returns
        -------
        kernel_parameters : dict
            Dictionary that maps each kernel parameter (name) to its index in kernel.theta, e.g.
            {'noise_level': 0, 'length_scale': [1,2,3,4]}

        """

        n_theta = len(kernel.theta)
        epsi = 1e-15
        kernel = kernel.clone_with_theta(np.arange(n_theta))
        expected = np.exp(kernel.theta)
        kernel_params = RbfAccelerator.get_kernel_params(kernel)

        def find_index_for_param(param_name):
            param = kernel_params[param_name]
            if (param_name != 'length_scale') or (isinstance(param, numbers.Number)):
                found = (param < expected * (1 + epsi)) & (param > expected * (1 - epsi))
                n_found = np.sum(found)
                if n_found > 1:
                    raise NotImplementedError("Cannot find theta mapping for parameter " + param_name)
                if n_found == 0:
                    return None
                return np.where(found)[0][0]

            params = kernel_params[param_name].reshape(-1)
            found = np.zeros(len(expected), dtype=bool)
            for param in params:
                tmp = (param < expected * (1 + epsi)) & (param > expected * (1 - epsi))
                if np.sum(tmp) == 1:
                    found[np.where(tmp)[0][0]] = 1
            n_found = np.sum(found)
            if n_found > params.size:
                raise NotImplementedError("Cannot find theta mapping for parameter " + param_name)
            if n_found == 0:
                return None
            ind = np.where(found)[0]
            assert (len(ind) == 1) or (set(np.unique(np.diff(ind))) == {1}), "Unsupported case of non-adjacent parameters for length scale"
            return ind

        theta_mapping = {param: find_index_for_param(param) for param in kernel_params}
        theta_mapping = {k: v for k, v in theta_mapping.items() if v is not None}
        num_params_found = sum(len(v) if type(v) is np.ndarray else 1 for k, v in theta_mapping.items())
        if num_params_found < len(kernel.theta):
            raise NotImplementedError("Cannot find the mapping for all kernel thetas")
        assert num_params_found == len(kernel.theta), "Unexpected case when mapping kernel thetas to parameters"

        return theta_mapping


def _ensure_cupy_array(x):
    """ Utility function that converts an input array to a cupy array through
    the 'from_dlpack' interface """
    if type(x) is cp.ndarray:
        return x
    return cp.from_dlpack(x)


# TODO: implent other kernels, e.g. Matern

# Cuda kernel text to calculate nearest neighbors with RBF kernels
_calc_nn_rbf_kernel_text = """
extern "C"
__global__ void calc_nn(
    DATA_TYPE const * ref,
    DATA_TYPE const * query,
    INDEX_TYPE *  k_indexes
#ifdef USE_GLOBAL_MEM_CACHE
    , DATA_TYPE * cache_distances,
    INDEX_TYPE * cache_indexes
#endif
)
{
    // Thread index
    INDEX_TYPE i_query = blockIdx.x * BLOCK_DIM + threadIdx.x;

    if (i_query >= N_QUERY)
        return;

#ifdef USE_STATIC_SHARED_MEM_CACHE
    __shared__ DATA_TYPE cache_distances[BLOCK_DIM][NUM_K_FOR_CACHE];
    __shared__ INDEX_TYPE cache_indexes[BLOCK_DIM][NUM_K_FOR_CACHE];
    DATA_TYPE * this_distances = cache_distances[threadIdx.x];
    INDEX_TYPE * this_indexes = cache_indexes[threadIdx.x];
#define CACHE_DIST_PITCH
#define CACHE_INDEX_PITCH

#elif defined(USE_DYNAMIC_SHARED_MEM_CACHE)
    extern __shared__ DATA_TYPE cache_distances[];
    INDEX_TYPE * cache_indexes = (INDEX_TYPE *) (cache_distances + BLOCK_DIM * NUM_K_FOR_CACHE);
    
#ifdef DYNAMIC_SHARED_MEM_ALT_ORDER
    DATA_TYPE * this_distances = cache_distances + threadIdx.x;
    INDEX_TYPE * this_indexes = cache_indexes + threadIdx.x;
#define CACHE_DIST_PITCH *BLOCK_DIM
#define CACHE_INDEX_PITCH *BLOCK_DIM
#else
    DATA_TYPE * this_distances = cache_distances + threadIdx.x * NUM_K_FOR_CACHE;
    INDEX_TYPE * this_indexes = cache_indexes + threadIdx.x * NUM_K_FOR_CACHE;
#define CACHE_DIST_PITCH
#define CACHE_INDEX_PITCH
#endif

#elif defined(USE_GLOBAL_MEM_CACHE)
    DATA_TYPE * this_distances = cache_distances + i_query;
    INDEX_TYPE * this_indexes = cache_indexes + i_query;
#endif

    DATA_TYPE const * query_point = query + i_query;

    for (SIZE_TYPE j=0; j<NUM_K; j++) this_indexes[j CACHE_INDEX_PITCH ] = -1;
    for (SIZE_TYPE j=0; j<NUM_K; j++) this_distances[j CACHE_DIST_PITCH ] = INFINITY;
    DATA_TYPE max_dist = INFINITY;

    for (SIZE_TYPE i_ref=0; i_ref<MAX_SEARCH_POINT; i_ref++) {
        DATA_TYPE distance = 0;
        DATA_TYPE const * ref_point = ref + i_ref REF_PITCH_POINT ;
        for (SIZE_TYPE f=0; f<NUM_FEATURE; f++) {
            DATA_TYPE tmp = ref_point[f REF_PITCH_FEATURE ] - query_point[f * QUERY_PITCH];
            distance += tmp * tmp;
        }

        // If this is within the k nearest, update it
        if (distance < max_dist) {
            SIZE_TYPE ii = NUM_K - 1;
            while (ii > 0) {
                if (this_distances[(ii - 1) CACHE_DIST_PITCH ] <= distance)
                    break;
                this_distances[ii CACHE_DIST_PITCH ] = this_distances[(ii - 1) CACHE_DIST_PITCH ];
                this_indexes[ii CACHE_INDEX_PITCH ] = this_indexes[(ii - 1) CACHE_INDEX_PITCH ];
                ii -= 1;
            }
            this_distances[ii CACHE_DIST_PITCH ] = distance;
            this_indexes[ii CACHE_INDEX_PITCH ] = (INDEX_TYPE) i_ref;
            max_dist = this_distances[(NUM_K - 1) CACHE_DIST_PITCH ];
        }
    }

    for (SIZE_TYPE i=0; i<min(NUM_K, MAX_SEARCH_POINT); i++) {
        k_indexes[i_query + i * INDEX_PITCH] = this_indexes[i CACHE_INDEX_PITCH ];
    }
}

"""

# Cuda kernel text to calculate the GP kernel and its gradient for RBF kernels
_fill_nn_RBF_kernel_text = """
#ifdef USE_F16
#include cuda_fp16.h
#endif

extern "C"
__global__ void fill_nn_kernel(
    DATA_TYPE const * data_points,
    NN_IND_DTYPE const * nn_points,
    DATA_TYPE const * rbf_length_scales,
    DATA_TYPE * nn_kernel,
    DATA_TYPE * nn_kernel_grad,
    DATA_TYPE const const_value,
    DATA_TYPE const white_value,
    DATA_TYPE const rbf_level
)
{
    INDEX_TYPE i_ref_point;
    // Thread index
    {
        size_t i_ref_size_t = blockIdx.x * BLOCK_DIM + threadIdx.x;
        if (i_ref_size_t >= NUM_REF_POINTS)
            return;
        i_ref_point = (INDEX_TYPE) i_ref_size_t;
    }
    NN_IND_DTYPE const * this_nn_points = nn_points + i_ref_point;
    DATA_TYPE * this_kernel = nn_kernel + i_ref_point;
    DATA_TYPE * this_kernel_grad = nn_kernel_grad + i_ref_point;
    DATA_TYPE tmp_dist[NUM_FEATURES], exp_value, kernel_value;

    for (INDEX_TYPE i_nn=0; i_nn<NUM_NN; i_nn++) {
        NN_IND_DTYPE const nn_i_index = *(this_nn_points + i_nn * PITCH_NN_POINTS);
        DATA_TYPE const * i_data_point = data_points + nn_i_index * PITCH_DATA_POINTS;
        for (INDEX_TYPE j_nn=0; j_nn<=i_nn; j_nn++) {
            NN_IND_DTYPE const nn_j_index = *(this_nn_points + j_nn * PITCH_NN_POINTS);
            DATA_TYPE const * j_data_point = data_points + nn_j_index * PITCH_DATA_POINTS;
            BIG_INDEX_TYPE ij_index = i_nn + j_nn * NUM_NN;
            BIG_INDEX_TYPE ji_index = j_nn + i_nn * NUM_NN;

            if ((nn_i_index == -1) || (nn_j_index == -1)) {
                kernel_value = (i_nn == j_nn) ? 1 : 0;
                if (WHITE_NOISE_INDEX > -1) {
                    this_kernel_grad[ij_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * WHITE_NOISE_INDEX] = 0;
                    this_kernel_grad[ji_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * WHITE_NOISE_INDEX] = 0;
                }
                if (CONSTANT_LEVEL_INDEX > -1) {
                    this_kernel_grad[ij_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * CONSTANT_LEVEL_INDEX] = 0;
                    this_kernel_grad[ji_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * CONSTANT_LEVEL_INDEX] = 0;
                }
                if (RBF_LEVEL_INDEX > -1) {
                    this_kernel_grad[ij_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * RBF_LEVEL_INDEX] = 0;
                    this_kernel_grad[ji_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * RBF_LEVEL_INDEX] = 0;
                }
                #ifdef IS_VEC_RBF_SCALE
                for (int i_scale=RBF_SCALE_VEC_INDEX_0; i_scale < RBF_SCALE_VEC_INDEX_1; i_scale++) {
                    this_kernel_grad[ij_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * i_scale] = 0;
                    this_kernel_grad[ji_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * i_scale] = 0;
                }
                #else
                if (RBF_SCALE_INDEX > -1) {
                    this_kernel_grad[ij_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * RBF_SCALE_INDEX] = 0;
                    this_kernel_grad[ji_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * RBF_SCALE_INDEX] = 0;
                }
                #endif
            } else {
                DATA_TYPE distance = 0;
                #ifdef IS_VEC_RBF_SCALE
                for (int i_feat=0; i_feat<NUM_FEATURES; i_feat++) {
                    tmp_dist[i_feat] = ((i_data_point[i_feat] - j_data_point[i_feat])*(i_data_point[i_feat] - j_data_point[i_feat]))/(
                        2.0f * rbf_length_scales[i_feat] * rbf_length_scales[i_feat]);
                    distance += tmp_dist[i_feat];
                }
                #else
                for (int i_feat=0; i_feat<NUM_FEATURES; i_feat++) {
                    tmp_dist[i_feat] = ((i_data_point[i_feat] - j_data_point[i_feat])*(i_data_point[i_feat] - j_data_point[i_feat]));
                    distance += tmp_dist[i_feat];
                }
                distance = distance / (2.0f * rbf_length_scales[0] * rbf_length_scales[0]);
                #endif
                exp_value = EXP(-distance) * rbf_level;
                kernel_value = exp_value + const_value;

                //if (distance == 0) {
                if (i_nn == j_nn) {
                    kernel_value += white_value;
                    if (WHITE_NOISE_INDEX > -1) {
                        this_kernel_grad[ij_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * WHITE_NOISE_INDEX] = white_value;
                        this_kernel_grad[ji_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * WHITE_NOISE_INDEX] = white_value;
                    }
                }
                if (CONSTANT_LEVEL_INDEX > -1) {
                    this_kernel_grad[ij_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * CONSTANT_LEVEL_INDEX] = const_value;
                    this_kernel_grad[ji_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * CONSTANT_LEVEL_INDEX] = const_value;
                }
                if (RBF_LEVEL_INDEX > -1) {
                    this_kernel_grad[ij_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * RBF_LEVEL_INDEX] = exp_value;
                    this_kernel_grad[ji_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * RBF_LEVEL_INDEX] = exp_value;
                }
                #ifdef IS_VEC_RBF_SCALE
                for (int i_scale=RBF_SCALE_VEC_INDEX_0; i_scale < RBF_SCALE_VEC_INDEX_1; i_scale++) {
                    int tmp = i_scale - RBF_SCALE_VEC_INDEX_0;
                    this_kernel_grad[ij_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * i_scale] = exp_value * 2 * tmp_dist[tmp];
                    this_kernel_grad[ji_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * i_scale] = exp_value * 2 * tmp_dist[tmp];
                }
                #else
                if (RBF_SCALE_INDEX > -1) {
                    this_kernel_grad[ij_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * RBF_SCALE_INDEX] = exp_value * 2 * distance;
                    this_kernel_grad[ji_index * PITCH_NN_KERNEL_GRAD + THETA_PITCH * RBF_SCALE_INDEX] = exp_value * 2 * distance;
                }
                #endif
            }
            this_kernel[ij_index * PITCH_NN_KERNEL] = kernel_value;
            this_kernel[ji_index * PITCH_NN_KERNEL] = kernel_value;
        }
    }
}
"""



import itertools
from abc import abstractmethod, ABC

import numpy as np
from sklearn import gaussian_process as skgp

from nngpr.base_nngpr import NngprWorker
from nngpr import batched_kernels
from nngpr.numpy_nngpr import NumpyProxy


class WorkerTester(ABC):

    __test__ = False

    def __init__(self, worker, *args, **kwargs):
        if type(worker) is not list:
            worker = [worker]
        self.workers = worker
        super().__init__(*args, **kwargs)

    def get_kern(self):
        return batched_kernels.RBF(length_scale=1.0)

    def test_empty(self):
        for worker, shape, dtype in itertools.product(self.workers, [2, (5, 6)], [np.float32, np.float64]):
            xp = worker.get_array_module()
            with self.subTest(worker=worker, shape=shape, dtype=dtype):
                xp.empty(shape, dtype=dtype)
            
    def test_get_array_module(self):
        for worker in self.workers:
            with self.subTest(worker=worker):
                xp = worker.get_array_module()
                x1 = xp.empty((5, 2, 3), dtype=np.float32)
                x2 = xp.empty((10, 2, 3), dtype=np.float32)
                self.assertTrue(xp.moveaxis(x1, 0, 2).shape == (2, 3, 5))
                self.assertTrue(xp.swapaxes(x2, 1, 2).shape == (10, 3, 2))
                xp.log(x1 + 2)
                xp.sqrt(x2)
                xp.arange(10)
                xp.where(x1 > 2)
                xp.sum(x1)
                xp.tile(x1, (3, 1, 1))
                x2 = xp.ones((10, 3, 4, 4), dtype=np.float32)
                x2 @ x2

    def test_to_float32(self):
        for worker in self.workers:
            with self.subTest(worker=worker):
                xp = worker.get_array_module()
                x1 = xp.ones((5, 2, 3), dtype=np.float32)
                x2 = xp.ones((10, 2, 3), dtype=np.float64)
                self.assertTrue(x1 is xp.to_float32(x1))
                self.assertFalse(x2 is xp.to_float32(x2))

    def test_to_numpy(self):
        for worker in self.workers:
            with self.subTest(worker=worker):
                xp = worker.get_array_module()
                x1 = xp.empty(3, dtype=np.float32)
                self.assertTrue(type(xp.to_numpy(x1)) is np.ndarray)

    def test_from_numpy(self):
        x = np.ones(6)
        for worker in self.workers:
            with self.subTest(worker=worker):
                xp = worker.get_array_module()
                x1 = xp.empty(3, dtype=np.float32)
                self.assertTrue(type(xp.from_numpy(x)) is type(x1))
        
    def test_find_nearest_neighbors(self):
        for worker, nn_type, ref_cut_index in itertools.product(self.workers, ['kernel-space', 'input-space'], [None, 0, 2, -1]):
            with self.subTest(worker=worker, nn_type=nn_type, ref_cut_index=ref_cut_index):
                xp = worker.get_array_module()
                ref = xp.ones((10, 3))
                query = xp.ones((4, 3)) + 2
                k = self.get_kern()
                res = worker.find_nearest_neighbors(ref, query, nn_type, k, 2, ref_cut_index=ref_cut_index)
                self.assertEqual(type(res), type(ref))

    def test_chofactor_and_chosolve(self):
        x = np.random.rand(3, 5, 5)
        x = x @ np.swapaxes(x, -1, -2) + np.eye(x.shape[-1])[None, ...]
        y = np.random.rand(3, 5, 8)
        for worker in self.workers:
            with self.subTest(worker=worker):
                xp = worker.get_array_module()
                L, L_lower = worker.batched_chofactor(xp.from_numpy(x))
                res = worker.batched_chosolve(L, xp.from_numpy(y), L_lower)
                self.assertEqual(type(res), type(xp.from_numpy(x)))
                self.assertTrue(np.allclose(xp.to_numpy(res), np.linalg.solve(x, y)))

    def test_fill_nn_kernel(self):
        for worker, eval_gradient in itertools.product(self.workers, [False, True]):
            with self.subTest(worker=worker, eval_gradient=eval_gradient):
                xp = worker.get_array_module()
                x = xp.zeros((10, 5), dtype=np.float32)
                nn_indices = np.random.randint(0, x.shape[0], (3, 4), dtype=np.int32)
                k = self.get_kern()
                res = worker.fill_nn_kernel(xp.from_numpy(x), xp.from_numpy(nn_indices), k, eval_gradient=eval_gradient)
                self.assertEqual(type(res[0]), type(x))
                self.assertEqual(type(res[1]), type(x))

    def test_random_normal(self):
        mu = np.ones((3, 1, 2, 1))
        sigma = np.ones((1, 5, 2, 1))
        shape = (3, 5, 2, 10)
        for worker in self.workers:
            with self.subTest(worker=worker):
                xp = worker.get_array_module()
                for random_state in [None, 0, 324]:
                    s = worker.check_random_state(None)
                    res = worker.random_normal(s, mu, sigma, shape)
                    self.assertEqual(type(res), type(xp.from_numpy(mu)))


class NngprTester:

    def __init__(self, *args, **kwargs):
        # Generate some fake data
        rng = np.random.RandomState(234058)
        n = 150
        nj = 4
        ny = 3
        self._x = rng.rand(n, nj)
        self._x_eval = rng.rand(50, nj)
        beta = rng.rand(self._x.shape[1], ny)
        beta[:, -1] += 1
        self._y = self._x @ beta + rng.rand(self._x.shape[0], ny)
        self._y[:, 0] *= 10
        super().__init__(*args, **kwargs)
        self.np = NumpyProxy()

    @abstractmethod
    def get_gpr(self, **kwargs):
        pass

    def get_kern(self, fixed_bounds: bool = False):
        """Helper function that returns the kernel"""
        if fixed_bounds:
            return batched_kernels.ConstantKernel(1, constant_value_bounds="fixed") * batched_kernels.RBF(
                length_scale=1, length_scale_bounds="fixed") + batched_kernels.WhiteKernel(
                1, noise_level_bounds="fixed")
        return batched_kernels.ConstantKernel(1) * batched_kernels.RBF(length_scale=1) + batched_kernels.WhiteKernel(1)


    def test_runs(self):
        """Tests that it runs with no errors"""

        def run(nn_type=None, distribute_method=None):
            gpr = self.get_gpr(nn_type=nn_type, distribute_method=distribute_method)
            if (len(gpr.workers) == 1) and (distribute_method is not None):
                return
            gpr.predict(self._x_eval, return_std=True)
            gpr.predict(self._x_eval, return_cov=True)
            gpr.sample_y(self._x_eval, n_samples=2)
            gpr.fit(self._x, self._y)
            gpr.predict(self._x)
            gpr.predict(self._x, return_std=True)
            gpr.predict(self._x_eval, return_std=True)
            gpr.predict(self._x, return_cov=True)
            gpr.predict(self._x_eval, return_cov=True)
            gpr.sample_y(self._x_eval, n_samples=2)

        for nn_type in ['kernel-space', 'input-space']:
            with self.subTest(nn_type=nn_type):
                run(nn_type=nn_type)

        for distribute_method in ['joblib', 'multiprocessing', 'multiprocessing-heterogenous', 'sequential']:
            with self.subTest(distribute_method=distribute_method):
                run(distribute_method=distribute_method)

    def test_same_lkl(self):
        """Tests that NNGPR gives the same likelihood as the scikit-learn Gaussian Process Regressor when the number of nearest neighbors
        is large enough"""
        kernel = self.get_kern()
        theta = np.ones(len(kernel.theta)) * 0.5
        for y_dim in [1, 100]:
            with self.subTest(y_dim=y_dim):
                mdl_ref = skgp.GaussianProcessRegressor(kernel=self.get_kern().bind_to_array_module(self.np), normalize_y=True)
                mdl_ref.X_train_, mdl_ref.y_train_, mdl_ref.kernel_ = self._x, self._y[:, :y_dim], self.get_kern().bind_to_array_module(self.np)
                lkl_ref, grad_ref = mdl_ref.log_marginal_likelihood(theta, eval_gradient=True)
                gpr = self.get_gpr(normalize_y=True, num_nn=self._x.shape[0])
                gpr.X_train_, gpr.y_train_, gpr.kernel_ = self._x, self._y[:, :y_dim], self.get_kern()
                lkl, grad = gpr.log_marginal_likelihood(theta, eval_gradient=True)

                self.assertEqual(grad.shape, grad.shape)
                delta = np.max(np.abs(lkl / lkl_ref - 1))
                self.assertLess(delta, 1e-10)
                delta = np.max(np.abs(grad / grad_ref - 1))
                self.assertLess(delta, 1e-10)

    def test_same_fit_results(self):
        """Tests that NNGPR.fit gives the same likelihood as the scikit-learn Gaussian Process Regressor fit when the number of nearest
        neighbors is large enough"""
        for y_dim in [1, 100]:
            with self.subTest(y_dim=y_dim):
                mdl_ref = skgp.GaussianProcessRegressor(kernel=self.get_kern().bind_to_array_module(self.np), normalize_y=True)
                mdl_ref.fit(self._x, self._y[:, :y_dim])
                gpr = self.get_gpr(
                    kernel=self.get_kern(), normalize_y=True, num_nn=self._x.shape[0])
                gpr.fit(self._x, self._y[:, :y_dim])
                delta = np.max(np.abs(mdl_ref.kernel_.theta / gpr.kernel_.theta - 1))
                self.assertLess(delta, 1e-9)

    def test_prior_predict_same_results(self):
        """Tests that NNGPR gives the same prior predictions as the scikit-learn Gaussian Process Regressor when the number of nearest
        neighbors is large enough"""
        mdl_ref = skgp.GaussianProcessRegressor(kernel=self.get_kern().bind_to_array_module(self.np), normalize_y=True)
        mean_ref, std_ref = mdl_ref.predict(self._x, return_std=True)
        _, cov_ref = mdl_ref.predict(self._x, return_cov=True)
        for conditioning_method in ['only-train', 'full', 'full-double-nn']:
            with self.subTest(conditioning_method=conditioning_method):
                gpr = self.get_gpr(
                    fixed_bounds=True, normalize_y=True,
                    num_nn=self._x.shape[0])
                _, cov = gpr.predict(self._x, return_cov=True, conditioning_method=conditioning_method)
                mean, std = gpr.predict(self._x, return_std=True, conditioning_method=conditioning_method)
                delta = mean - mean_ref
                self.assertLess(np.max(np.abs(delta)), 1e-10)
                delta = std - std_ref
                self.assertLess(np.max(np.abs(delta)), 1e-10)
                if conditioning_method != 'only-train':
                    delta = cov - cov_ref
                    self.assertLess(np.max(np.abs(delta)), 1e-10)

    def test_predict_same_results(self):
        """Tests that NNGPR gives the same posterior predictions as the scikit-learn Gaussian Process Regressor when the number of nearest
        neighbors is large enough"""
        for conditioning_method, use_eval, y_dim in itertools.product(
                ['only-train', 'full', 'full-double-nn'], [True, False], [1, 100]):
            with self.subTest(conditioning_method=conditioning_method, use_eval=use_eval, y_dim=y_dim):
                x_eval = self._x_eval if use_eval else self._x

                mdl_ref = skgp.GaussianProcessRegressor(kernel=self.get_kern(fixed_bounds=True).bind_to_array_module(self.np), normalize_y=True)
                mdl_ref.fit(self._x, self._y[:, :y_dim])
                mean_ref, std_ref = mdl_ref.predict(x_eval, return_std=True)
                _, cov_ref = mdl_ref.predict(x_eval, return_cov=True)

                gpr = self.get_gpr(
                    fixed_bounds=True, normalize_y=True,
                    num_nn=self._x.shape[0] + x_eval.shape[0] if conditioning_method == 'full' else self._x.shape[0])
                gpr.fit(self._x, self._y[:, :y_dim])
                _, cov = gpr.predict(x_eval, return_cov=True, conditioning_method=conditioning_method)
                mean, std = gpr.predict(x_eval, return_std=True, conditioning_method=conditioning_method)

                self.assertEqual(mean.shape, mean_ref.shape)
                self.assertEqual(std.shape, std_ref.shape)
                self.assertEqual(cov.shape, cov_ref.shape)
                delta = mean - mean_ref
                self.assertLess(np.max(np.abs(delta)), 1e-10)
                delta = std - std_ref
                self.assertLess(np.max(np.abs(delta)), 1e-10)
                if conditioning_method != 'only-train':
                    delta = cov - cov_ref
                    self.assertLess(np.max(np.abs(delta)), 1e-10)

    def test_prior_sample(self):
        """Tests that NNGPR prior samples are consistent with predictions and with the prior samples from the scikit-learn Gaussian
        Process Regressor when the number of nearest neighbors is large enough"""
        n_samples = 10000
        random_state = 1
        for conditioning_method in ['only-train', 'full', 'full-double-nn']:
            with self.subTest(conditioning_method=conditioning_method):
                gpr = self.get_gpr(
                    fixed_bounds=True, normalize_y=True,
                    num_nn=self._x.shape[0])
                samples = gpr.sample_y(self._x, n_samples=n_samples, random_state=random_state,
                                       conditioning_method=conditioning_method)
                mean_pred, std_pred = gpr.predict(
                    self._x, conditioning_method=conditioning_method, return_std=True)
                mdl_ref = skgp.GaussianProcessRegressor(kernel=self.get_kern(fixed_bounds=True).bind_to_array_module(self.np), normalize_y=True)
                samples_ref = mdl_ref.sample_y(self._x, n_samples=n_samples, random_state=random_state)
                mean_ref, std_ref = mdl_ref.predict(self._x, return_std=True)
                random_state += 1
                mean_delta = np.max(np.abs(np.mean(samples_ref, axis=1) - mean_ref))
                self.assertLess(np.max(np.abs(np.mean(samples, axis=1) - mean_pred)), mean_delta * 2)
                std_delta = np.max(np.abs(np.std(samples_ref, axis=1) - std_ref))
                self.assertLess(np.max(np.abs(np.std(samples, axis=1) - std_pred)), std_delta * 2)

    def test_sample(self):
        """Tests that NNGPR posterior samples are consistent with predictions and with the posterior samples from the scikit-learn Gaussian
        Process Regressor when the number of nearest neighbors is large enough"""
        n_samples = 1000 # int(2 * 1024**3 / self._x.shape[0] / 8) #10000
        random_state = 1
        for conditioning_method, use_eval, y_dim in itertools.product(
                ['only-train', 'full', 'full-double-nn'], [False, True], [1, 100]):
            with self.subTest(
                    conditioning_method=conditioning_method, use_eval=True, y_dim=y_dim):
                x_eval = self._x_eval if use_eval else self._x
                gpr = self.get_gpr(
                    fixed_bounds=True, normalize_y=True,
                    num_nn=self._x.shape[0])
                gpr.fit(self._x, self._y[:, :y_dim])
                samples = gpr.sample_y(x_eval, n_samples=n_samples, random_state=random_state,
                                       conditioning_method=conditioning_method)
                mean_pred, std_pred = gpr.predict(
                    x_eval, conditioning_method=conditioning_method, return_std=True)
                mdl_ref = skgp.GaussianProcessRegressor(kernel=self.get_kern(fixed_bounds=True).bind_to_array_module(self.np), normalize_y=True)
                mdl_ref.fit(self._x, self._y[:, :y_dim])
                samples_ref = mdl_ref.sample_y(x_eval, n_samples=n_samples, random_state=random_state)
                mean_ref, std_ref = mdl_ref.predict(x_eval, return_std=True)
                random_state += 1

                self.assertEqual(samples.shape, samples_ref.shape)
                mean_delta = np.max(np.abs(np.mean(samples_ref, axis=-1) - mean_ref))
                self.assertLess(np.max(np.abs(np.mean(samples, axis=-1) - mean_pred)), mean_delta * 2)
                std_delta = np.max(np.abs(np.std(samples_ref, axis=-1) - std_ref))
                self.assertLess(np.max(np.abs(np.std(samples, axis=-1) - std_pred)), std_delta * 2)

    def test_large_sample_size(self):
        """Tests that Nngpr sample_y works on large sizes"""
        
        size_gb = 2
        random_state = 1
        x = self._x
        gpr = self.get_gpr(fixed_bounds=True, normalize_y=True, num_nn=32)
        n_samples = int(size_gb * 1024**3 / self._x.shape[0] / 8)
        samples = gpr.sample_y(x, n_samples=n_samples, random_state=random_state, 
                               conditioning_method='only-train')
        self.assertEqual(samples.shape, (x.shape[0], n_samples))

        random_state = 2
        x = np.concatenate([self._x] * int(n_samples / self._x.shape[0] / 10), axis=0)
        gpr = self.get_gpr(fixed_bounds=True, normalize_y=True, num_nn=32)
        n_samples = int(2 * 1024**3 / x.shape[0] / 8)
        samples = gpr.sample_y(x, n_samples=n_samples, random_state=random_state, 
                               conditioning_method='only-train')
        self.assertEqual(samples.shape, (x.shape[0], n_samples))
        